/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
/************************************************************************
** MODULE INFORMATION*
**********************
**     FILE     NAME:       nwtcp.c
**     SYSTEM   NAME:       nw
**     ORIGINAL AUTHOR(S):  Jan van Oorschot
**     VERSION  NUMBER:     1.00
**     CREATION DATE:       24 march 1993
**
** DESCRIPTION: TCP interface for the Network library
**              This module creates TCP sockets and uses
**		the blocking library to be notified when
**		something happens on the socket.
**		The NW layer will be notified by
**		this layer when frames are received and
**		connections are opened/closed.
*************************************************************************
** CHANGE INFORMATION **
*************************
** REVISION:    $Revision$
** WORKFILE:    $Workfile$
** LOGINFO:     $Log$
*************************************************************************/
#include 	<stdio.h>
#include 	<stdlib.h>
#include	<string.h>
#include	<dnpap.h>
#include	<config.h>
#include	<block.h>
#include    <dinet.h>
#include	"../nw.h"
#include	"nwtcp.h"

/* structure containing pointers to the current connections */
TCPCONN *TCPList;

/* static variables */
static	int	ServerSocket;
static	BLOCK	ServerBlock;
static struct sockaddr_in tcp_server;

TCPPARAM TCPDefaultParam =
{
    1,				/* use the server port */
    2001			/* default server port */
};

/* status vars passed by the upper layer */
static int (*TCPCBOpen)(NWCONN *,char *);
static int (*TCPCBClose)(NWCONN *);
static int (*TCPCBDgramRec)(NWCONN *,BYTE *,int); /* never used */
static TCPPARAM *TCPParam;
static char TCPEntry[32];

/* local procedures prototypes*/
static int	CBTcpServerBlock(BLOCK *blk, int reason,void *param);
static int	CBTcpSocketBlock(BLOCK *blk, int reason,void *param);
static int 	ConnectRequest(void);
    
/************************************************************************
** NAME:        TCPInit
** SYNOPSIS:    int TCPInit(
**			int (*CBOpen)(NWCONN *,char *),
**			int (*CBClose)(NWCONN *),
**			int (*CBDgramRec)(NWCONN *,BYTE *,int),
**                      void *param);
**              
** DESCRIPTION: Initialises the TCP iface. The callback
**		functions will be called by the TCP iface
**		layer when a connection is opened/closed
**		by an other network node.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int TCPInit(
	    int (*CBOpen)(NWCONN *,char *),
	    int (*CBClose)(NWCONN *),
	    int (*CBDgramRec)(NWCONN *,BYTE *,int),
            void *param)
{
    SHORT shrt;
    char tbuf[256];
    int iret;
    struct hostent *hostnm;

    if ( sock_init() != 0)
	return(NW_ERROR+2);

    TCPCBOpen = CBOpen;
    TCPCBClose = CBClose;
    TCPCBDgramRec = CBDgramRec;		/* never used */
    if(param != NULL)
	TCPParam = param;
    else
    {
	/* try to find values from the configuration file */
	TCPParam = &TCPDefaultParam;
	if(ConfigGetShort("tcp.serverport",&shrt))
	    TCPParam->ServerPort = (int) &shrt;
	if(ConfigGetShort("tcp.server",&shrt))
	    TCPParam->server = 1;  /* use the server port */
	if(ConfigGetShort("tcp.noserver",&shrt))
	    TCPParam->server = 0;  /* use the server port */
    }

    TCPList = NULL;  /* no current connections */

    if(TCPParam->server != 0)
    {
	    /* create the server socket */
	if ((ServerSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	    DnpapMessage(DMC_WARNING,NW_ERROR+3,"can't create server socket");
	    return NW_ERROR+3;
	}
	tcp_server.sin_family = AF_INET;
	tcp_server.sin_port   = htons(TCPParam->ServerPort);
	tcp_server.sin_addr.s_addr = INADDR_ANY;
	if (bind(ServerSocket,
		 (struct sockaddr *) &tcp_server, 
		 sizeof(tcp_server)) < 0)
	{
	    DnpapMessage(DMC_WARNING,NW_ERROR+4,"can't bind server socket");
	    return NW_ERROR+4;
	}
	if (listen(ServerSocket, 5) != 0)
	{
	    DnpapMessage(DMC_WARNING,NW_ERROR+5,
			 "can't listen to server socket");
	    return NW_ERROR+5;
	}

	/* create the block on the server port */
	ServerBlock.sckt = ServerSocket;
	if( (iret = BlockRegister(
	    &ServerBlock,
	    &ServerBlock,
	    CBTcpServerBlock)) != 0)
	{
	    DnpapMessage(DMC_WARNING,NW_ERROR+6,
			 "failed to register server block");
	    return NW_ERROR+6;
	}
    }
    /* remember this entry for reporting purposes */
    strcpy(TCPEntry,"tcp:");
    gethostname(tbuf,256);
    hostnm = gethostbyname(tbuf);
    strcat(TCPEntry,Inet_NtoA(hostnm->h_addr));
    strcat(TCPEntry,":");
    sprintf(tbuf,"%03d",ntohs(tcp_server.sin_port));
    strcat(TCPEntry,tbuf);

    return 0;
}

/************************************************************************
** NAME:        TCPExit
** SYNOPSIS:    int TCPExit(void)
**              
** DESCRIPTION: Terminates the TCP layer.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int TCPExit(void)
{
    TCPCONN *ptr,*cptr;
    /* close the server socket */
    if(TCPParam->server != 0)
    {
	BlockDeRegister(&ServerBlock);
	soclose(ServerSocket);
    }
    /* close all open TCP connections */
    ptr = TCPList;
    while (ptr != NULL)
    {
	cptr = ptr;
	ptr = ptr->next;
	TCPClose((NWCONN *) cptr);
    }
    return 0;
}
/************************************************************************
** NAME:        TCPOpen
** SYNOPSIS:    NWCONN *TCPOpen(
**			char *ConnInfo,
**			int (*CBReceive)(NWCONN *,BYTE *,len));
**              
** DESCRIPTION: Opens a tcp connection and returns the 
**		connection pointer.
**		The info needed to build the TCP connection is
**		in <ConnInfo>.
**		The callback function <CBReceive> will be
**		called when a frame is received through this
**		connection.
** RETURNS:     NULL	can't open connection
**              else    connection structure
*************************************************************************/
NWCONN *TCPOpen(char *ConnInfo,int (*CBReceive)(NWCONN *,BYTE *,int))
{
    TCPCONN *con;
    BLOCK *blk;
    char *cptr;
    int iret;
    struct hostent *hostnm;
    unsigned long inaddr;
    char tbuf[256];

    /* allocate and the structures needed for a connection */
    if((con = DnpapMalloc(sizeof(TCPCONN)))==NULL)
	return NULL;
    if((con->nwinfo.ConnInfo = DnpapMalloc(32)) == NULL)
    {
	DnpapFree(con);
	return NULL;
    }
    if((blk = DnpapMalloc(sizeof(BLOCK))) == NULL)
    {
	DnpapFree(con->nwinfo.ConnInfo);
	DnpapFree(con);
	return NULL;
    }

    con->tcpinfo.CBRec = CBReceive;
	
    /* create the TCP socket */
    strcpy(tbuf,ConnInfo);
    cptr = tbuf;
    while(*cptr != ':' && *cptr != '\0') cptr++;
    if(*cptr == '\0')
    {
	DnpapFree(con->nwinfo.ConnInfo);
	DnpapFree(con);
	DnpapFree(blk);
	DnpapMessage(DMC_WARNING,NW_ERROR+20,"no port number to connect to");
	return NULL;
    }
    *cptr = '\0';
    con->tcpinfo.conPort = atoi(cptr+1);
    /* create socket, check for consistency */
    if ((con->tcpinfo.sckt = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
	DnpapFree(con->nwinfo.ConnInfo);
	DnpapFree(con);
	DnpapFree(blk);
	DnpapMessage(DMC_WARNING,NW_ERROR+23,"can't create socket");
	return NULL;
    }
    /* read message-discard mode */
    ioctl(con->tcpinfo.sckt,I_SRDOPT,(caddr_t)RMSGD); 
    /* determine address, first dotted, then fullname */
    inaddr = inet_addr(tbuf);
    if((hostnm = gethostbyaddr((char *) &inaddr,4,2)) == NULL)
    {
	/* normal hostname */
	hostnm = gethostbyname(tbuf);
	if (hostnm == NULL)
	{
	    soclose(con->tcpinfo.sckt);
	    DnpapFree(con->nwinfo.ConnInfo);
	    DnpapFree(con);
	    DnpapFree(blk);
	    DnpapMessage(DMC_WARNING,NW_ERROR+21,"gethostbyname failed");
	    return NULL;
	}
    }
    con->tcpinfo.addr.sin_addr.s_addr =  /* ulong */
	*((unsigned long *)(hostnm->h_addr));
    con->tcpinfo.addr.sin_family      = AF_INET;
    con->tcpinfo.addr.sin_port        = htons(con->tcpinfo.conPort);
    
    /* finaly, do the connect */
    if (connect(con->tcpinfo.sckt, 
		(struct sockaddr *) &(con->tcpinfo.addr), 
		sizeof(con->tcpinfo.addr)) < 0)
    {
	soclose(con->tcpinfo.sckt);
	DnpapFree(con->nwinfo.ConnInfo);
	DnpapFree(con);
	DnpapFree(blk);
	DnpapMessage(DMC_WARNING,NW_ERROR+23,"can't connect");
	return NULL;
    }
    /* create the BLOCK and attach block and connection*/
    blk->sckt = con->tcpinfo.sckt;
    con->tcpinfo.blk = blk;
    if( (iret = BlockRegister(blk,con,CBTcpSocketBlock)) != 0)
    {
	soclose(con->tcpinfo.sckt);
	DnpapFree(con->nwinfo.ConnInfo);
	DnpapFree(con);
	DnpapFree(blk);
	DnpapMessage(DMC_WARNING,NW_ERROR+24,"can't register socket block");
	return NULL;
    }

    /* termine a correct connection string */
    con->nwinfo.ConnInfo = DnpapMalloc(32);
    strcpy(con->nwinfo.ConnInfo,"tcp:");
    strcat(con->nwinfo.ConnInfo,Inet_NtoA(con->tcpinfo.addr.sin_addr));
    strcat(con->nwinfo.ConnInfo,":");
    sprintf(tbuf,"%03d",ntohs(con->tcpinfo.addr.sin_port));
    strcat(con->nwinfo.ConnInfo,tbuf);

    /* put new connection in the list */
    con->prior = NULL;
    con->next = TCPList;
    if(con->next != NULL)
	con->next->prior = con;
    TCPList = con;

    return (NWCONN *) con;
}
/************************************************************************
** NAME:        TCPClose
** SYNOPSIS:    int TCPClose(NWCONN *conn);
**              
** DESCRIPTION: Closes a TCP connection.
** RETURNS:     0 -->   no error
**              else    error code
*************************************************************************/
int TCPClose(NWCONN *conn)
{
    TCPCONN	*tcon;
    tcon = (TCPCONN *) conn;

    /* find and remove the connection in the list */
    if(TCPList == NULL)
	return NW_ERROR+1;
    if(TCPList == tcon)
    {
	TCPList = TCPList->next;
	if(TCPList!=NULL)
	    TCPList->prior = NULL;
    } else {
	tcon->prior->next = tcon->next;
	tcon->next->prior = tcon->prior;
    }

    /* notify NW of close */
    TCPCBClose(conn);
    /* deregister and free */
    BlockDeRegister(tcon->tcpinfo.blk);
    soclose(tcon->tcpinfo.sckt);
    DnpapFree(tcon->tcpinfo.blk);
    DnpapFree(tcon->nwinfo.ConnInfo);
    DnpapFree(tcon);
    return 0;
}
/************************************************************************
** NAME:        TCPSend
** SYNOPSIS:    int TCPSend(NWCONN *conn,BYTE *frame, int len)
**              
** DESCRIPTION: Send a frame through the TCP connection.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
int TCPSend(NWCONN *conn,BYTE *frame,int len)
{
    TCPCONN *tcon;
    int iret;
    tcon = (TCPCONN *) conn;
    iret = send(tcon->tcpinfo.sckt,frame,len,0);
    
    if(iret < 0 || iret != len)
    	return 1;
    else 
    	return 0;
}
/************************************************************************
** NAME:        TCPSetRecCB
** SYNOPSIS:    void TCPSetRecCB(
**			NWCONN *conn,
**		      	int (*CBReceive)(NWCONN *con,BYTE *frame,int len))
**              
** DESCRIPTION: Sets the receive callback function a a connection.
**		should be called for the Open Callback of the application.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
void TCPSetRecCB(
	NWCONN *conn,
      	int (*CBReceive)(NWCONN *con,BYTE *frame,int len))
{
    TCPCONN *tcon;
    tcon = (TCPCONN *) conn;
    tcon->tcpinfo.CBRec = CBReceive;
}
/************************************************************************
** NAME:        TCPGetEntry
** SYNOPSIS:    char *TCPGetEntry()
** DESCRIPTION: Report the connection entry of this layer. Should
**		be in such a format that other node can use 
**		this entry to connect to this iface.
** RETURNS:     0 -->   no error
**              else    error code
** RETURNS:     void
*************************************************************************/
char *TCPGetEntry(void)
{
    return TCPEntry;
}

/************************************************************************
** NAME:        TCPConfig
** SYNOPSIS:    int TCPConfig(char *cfg);
**              
** DESCRIPTION: Configs a TCP connection.
**		The TCP library understands the following configuration
**		options:
**
**		noserver	dont start serverport
**		server		do start serverport
**		serverport=<nr> sets the TCP serverport
**		
** RETURNS:     0 -->   no error
**              else    error code
*************************************************************************/
int TCPConfig(char *cfg)
{
	char cmd[32];
	char *ptr,*arg,*cptr;
	
	ptr = cfg;
	cptr = cmd;
	while((*ptr != '=') && (*ptr != '\0'))
		*cptr++ = *ptr++;
	
	*cptr='\0';
	if(*ptr == '\0')
		arg=NULL;
	else
		arg=ptr+1;
		
	/* cmd should now contain the command, arg should point
	   to the argument string */
	if(strcmp(cmd,"server") == 0)
	{
		/* start server at init time */
		TCPDefaultParam.server=1;
	}
	else if(strcmp(cmd,"noserver") == 0)
	{
		/* do not start server */
		TCPDefaultParam.server=0;
	}
	else if(strcmp(cmd,"serverport") == 0)
	{
		/* set serverport */
		if(arg!=NULL)
			TCPDefaultParam.ServerPort=atoi(arg);
	}
	else
	{
		/* unknown command */
		return NW_ERROR;
	}
	return 0;
}
static int	CBTcpServerBlock(BLOCK *blk, int reason,void *param)
{
    if(blk != &ServerBlock)
    {
	DnpapMessage(DMC_WARNING,NW_ERROR+7,"wrong server block");
	return NW_ERROR+7;
    }
    switch(reason)
    {
    case BLOCKEVENT_READ:
	/* on the server socket, this is interpreted as a connection */
	/* request, this is handle by a seperate function */
	return ConnectRequest();
	break;
    case BLOCKEVENT_WRITE:
	/* ignore these events */
	DnpapMessage(DMC_WARNING,NW_ERROR+30,"BLOCKEVENT_WRITE on serverport");
	break;
    case BLOCKEVENT_EXCEPT:
	/* ignore these events */
	DnpapMessage(DMC_WARNING,NW_ERROR+30,"BLOCKEVENT_EXCEPTon serverport");
	break;
    default:
	DnpapMessage(DMC_WARNING,NW_ERROR+8,"unknown block event");
	return NW_ERROR+8;
	break;
    }
    return 0;
}

static BYTE RecBuf[10000];
static int	CBTcpSocketBlock(BLOCK *blk, int reason, void *param)
{
    TCPCONN	*tconn;
    int nrrec;

    /* remember, param contains a pointer to a TCPCONN */
    tconn = param;
    switch(reason)
    {
    case BLOCKEVENT_READ:
	/* a read of length 0 is interpreted as a close */
	nrrec = read(tconn->tcpinfo.sckt,RecBuf,10000);
	if(nrrec < 0)
	{
	    DnpapMessage(DMC_WARNING,NW_ERROR+42,"read error");
	    return NW_ERROR+42;
	}
	if(nrrec == 0)
	{
	    /* this is interpreted as a close of the socket */
	    return TCPClose((NWCONN *) tconn);
	}
	/* notify NW of the receiped */
	if(tconn->tcpinfo.CBRec != NULL)
	    return (tconn->tcpinfo.CBRec)((NWCONN *) tconn,RecBuf,nrrec);
	break;
    case BLOCKEVENT_WRITE:
	break;
    case BLOCKEVENT_EXCEPT:
	DnpapMessage(DMC_WARNING,NW_ERROR+30,"EXEPTEVENT_WRITE on normalport");
	break;
    default:
	DnpapMessage(DMC_WARNING,NW_ERROR+8,"unknown block event");
	return NW_ERROR+8;
	break;
    }
    return 0;
}

static int ConnectRequest(void)
{
    int nsckt,iret;
    char tbuf[8];
    TCPCONN *tcon;
    BLOCK   *blk;
    int ts;
    /* someone has requested a connection */
    /* allocate a connection and block structure */
    if((tcon = DnpapMalloc(sizeof(TCPCONN))) == NULL )
    {
	DnpapMessage(DMC_WARNING,NW_ERROR+40,"alloc failed");
	return NW_ERROR+40;
    }
    if((blk = DnpapMalloc(sizeof(BLOCK))) == NULL )
    {
	DnpapFree(tcon);
	DnpapMessage(DMC_WARNING,NW_ERROR+40,"alloc failed");
	return NW_ERROR+40;
    }
    /* nsckt will become a new socket, fully resolved, 
       the original ServerSocket will stay open
       for future connection requests */
    ts = sizeof(tcon->tcpinfo.addr);
    nsckt = accept(
		   ServerSocket,
		   (struct sockaddr *)&(tcon->tcpinfo.addr),
		   &ts);

      /* read message-discard mode */
    if(nsckt < 0)
    {
	DnpapFree(blk);
	DnpapFree(tcon);
	DnpapMessage(DMC_WARNING,NW_ERROR+41,"accept failed");
	return NW_ERROR+41;
    }
    ioctl(nsckt,I_SRDOPT,(caddr_t)RMSGD); 	
    
    /* termine a correct connection string */
    tcon->nwinfo.ConnInfo = DnpapMalloc(32);
    strcpy(tcon->nwinfo.ConnInfo,"tcp:");
    strcat(tcon->nwinfo.ConnInfo,Inet_NtoA(tcon->tcpinfo.addr.sin_addr));
    strcat(tcon->nwinfo.ConnInfo,":");
    sprintf(tbuf,"%03d",ntohs(tcon->tcpinfo.addr.sin_port));
    strcat(tcon->nwinfo.ConnInfo,tbuf);
    
    tcon->tcpinfo.sckt = nsckt;
    tcon->tcpinfo.blk = blk;
    tcon->tcpinfo.conPort = 0;
    tcon->tcpinfo.CBRec = NULL;		/* should be set by NW layer */
    /* inform NW layer of creation */
    TCPCBOpen((NWCONN *)tcon,tcon->nwinfo.ConnInfo);
    /* after this, the read-callback should be known, and
       the block can be attached */
    blk->sckt = nsckt;
    if((iret = BlockRegister(blk,tcon,CBTcpSocketBlock)) != 0)
    {
	DnpapFree(blk);
	DnpapFree(tcon->nwinfo.ConnInfo);
	DnpapFree(tcon);
	DnpapMessage(DMC_WARNING,NW_ERROR+31,"could not register block!");
	return NW_ERROR+31;
    }
    /* connect the new connection to the connection list */
    tcon->next = TCPList;
    tcon->prior = NULL;
    TCPList->prior = tcon;
    TCPList = tcon;
    return 0;
}
